From 40364a9f3926a9ebc8cad4534ab2221e2a1d2574 Mon Sep 17 00:00:00 2001
From: jian he <jian.universality@gmail.com>
Date: Wed, 5 Mar 2025 20:38:27 +0800
Subject: [PATCH v2 2/3] fast default for domain with constraints

This is primarily done by evaluating CoerceToDomain with soft error support.

If we evaluate CoerceToDomain to false, it means in ATExecAddColumn,
the defval node evaluation value cannot be cast to the domain.
However, we cannot fail at the Phase 2 stage in cases where the table is empty.

Therefore, if an error occurred while evaluation, do not raise the error,
we signal Phase 3 to do table rewrite, error will be raised on Phase 3.

Thanks to commit aaaf9449ec6be62cb0d30ed3588dc384f56274bf[1],
ExprState.escontext (ErrorSaveContext) was added, and ExecEvalConstraintNotNull,
ExecEvalConstraintCheck were changed to use errsave instead of hard error.
Now we can evaluate CoerceToDomain in a soft error way.

NOTE: this patch does not consider domain with volatile check constraints.

discussion: https://postgr.es/m/CACJufxE_+iZBR1i49k_AHigppPwLTJi6km8NOsC7FWvKdEmmXg@mail.gmail.com

[1]: https://git.postgresql.org/cgit/postgresql.git/commit/?id=aaaf9449ec6be62cb0d30ed3588dc384f56274bf
---
 src/backend/commands/tablecmds.c           | 43 ++++++++--
 src/test/regress/expected/fast_default.out | 91 ++++++++++++++++++++++
 src/test/regress/sql/fast_default.sql      | 57 ++++++++++++++
 3 files changed, 186 insertions(+), 5 deletions(-)

diff --git a/src/backend/commands/tablecmds.c b/src/backend/commands/tablecmds.c
index 8826ca5c32c..f7c8348c7ff 100644
--- a/src/backend/commands/tablecmds.c
+++ b/src/backend/commands/tablecmds.c
@@ -7331,6 +7331,7 @@ ATExecAddColumn(List **wqueue, AlteredTableInfo *tab, Relation rel,
 	 * NULL if so, so without any modification of the tuple data we will get
 	 * the effect of NULL values in the new column.
 	 *
+	 * XXXX this para should be removed?
 	 * An exception occurs when the new column is of a domain type: the domain
 	 * might have a not-null constraint, or a check constraint that indirectly
 	 * rejects nulls.  If there are any domain constraints then we construct
@@ -7358,6 +7359,7 @@ ATExecAddColumn(List **wqueue, AlteredTableInfo *tab, Relation rel,
 	{
 		bool		has_domain_constraints;
 		bool		has_missing = false;
+		bool		no_explicit_defval = false;
 
 		/*
 		 * For an identity column, we can't use build_column_default(),
@@ -7375,6 +7377,21 @@ ATExecAddColumn(List **wqueue, AlteredTableInfo *tab, Relation rel,
 		else
 			defval = (Expr *) build_column_default(rel, attribute->attnum);
 
+		/*
+		 * if defval is there, atthasdef is false, that means the defval comes
+		 * from domain default value, no explicit DEFAULT expression specified.
+		 * In that case we need evaluate defval error safe way, so
+		 * domain specification such as ``check(value > 10) default 8 ``
+		 * can be added to empty table.
+		*/
+		if (defval)
+		{
+			TupleDesc	rd_att = rel->rd_att;
+			Form_pg_attribute att_tup = TupleDescAttr(rd_att, attribute->attnum - 1);
+			if (!att_tup->atthasdef)
+				no_explicit_defval = true;
+		}
+
 		/* Build CoerceToDomain(NULL) expression if needed */
 		has_domain_constraints = DomainHasConstraints(attribute->atttypid);
 		if (!defval && has_domain_constraints)
@@ -7397,6 +7414,12 @@ ATExecAddColumn(List **wqueue, AlteredTableInfo *tab, Relation rel,
 													-1);
 			if (defval == NULL) /* should not happen */
 				elog(ERROR, "failed to coerce base type to domain");
+			/*
+			 * if domain have not-null constraint or check constraint
+			 * equivalent to not-null, we only want it failure when table have
+			 * some rows. so do it in soft error way.
+			*/
+			no_explicit_defval = true;
 		}
 
 		if (defval)
@@ -7419,13 +7442,13 @@ ATExecAddColumn(List **wqueue, AlteredTableInfo *tab, Relation rel,
 			 * specified DEFAULT value outside of the heap.  This is only
 			 * allowed for plain relations and non-generated columns, and the
 			 * default expression can't be volatile (stable is OK).  Note that
-			 * contain_volatile_functions deems CoerceToDomain immutable, but
-			 * here we consider that coercion to a domain with constraints is
-			 * volatile; else it might fail even when the table is empty.
+			 * contain_volatile_functions deems CoerceToDomain immutable.
+			 * We do support soft error evaluation of CoerceToDomain if
+			 * ExprState->escontext is not NULL. In that case if evaluation failed,
+			 * set table rewrite to true, let's fail on Phase 3.
 			 */
 			if (rel->rd_rel->relkind == RELKIND_RELATION &&
 				!colDef->generated &&
-				!has_domain_constraints &&
 				!contain_volatile_functions((Node *) defval))
 			{
 				EState	   *estate;
@@ -7435,10 +7458,20 @@ ATExecAddColumn(List **wqueue, AlteredTableInfo *tab, Relation rel,
 
 				/* Evaluate the default expression */
 				estate = CreateExecutorState();
-				exprState = ExecPrepareExpr(defval, estate);
+				if (!no_explicit_defval)
+					exprState = ExecPrepareExpr(defval, estate);
+				else
+					exprState = ExecPrepareExprSafe(defval, estate);
+
 				missingval = ExecEvalExpr(exprState,
 										  GetPerTupleExprContext(estate),
 										  &missingIsNull);
+
+				if (SOFT_ERROR_OCCURRED(exprState->escontext))
+				{
+					missingIsNull = true;
+					tab->rewrite |= AT_REWRITE_DEFAULT_VAL;
+				}
 				/* If it turns out NULL, nothing to do; else store it */
 				if (!missingIsNull)
 				{
diff --git a/src/test/regress/expected/fast_default.out b/src/test/regress/expected/fast_default.out
index ccbcdf8403f..1ab610a1410 100644
--- a/src/test/regress/expected/fast_default.out
+++ b/src/test/regress/expected/fast_default.out
@@ -317,11 +317,102 @@ SELECT a, b, length(c) = 3 as c_ok, d, e >= 10 as e_ok FROM t2;
  2 | 3 | t    | {This,is,abcd,the,real,world} | t
 (2 rows)
 
+---test fast default over domains with check constraint
+CREATE DOMAIN domain5 AS int check(value > 10) default 8;
+create domain domain6 as int not null;
+create domain domain7 as int check(value is not null);
+create domain domain8 as int check(value > 10) DEFAULT random(min=>10, max=>100);
+create domain domain9 as int check(value is not null) default 9;
+CREATE TABLE t3(a int);
+ALTER TABLE t3 ADD COLUMN b domain5 default 1; --error
+ERROR:  value for domain domain5 violates check constraint "domain5_check"
+ALTER TABLE t3 ADD COLUMN b domain6 default NULL; --error
+ERROR:  domain domain6 does not allow null values
+ALTER TABLE t3 ADD COLUMN b domain6 default 11 + NULL; --error
+ERROR:  domain domain6 does not allow null values
+ALTER TABLE t3 ADD COLUMN b domain7 default 11 + NULL; --error
+ERROR:  value for domain domain7 violates check constraint "domain7_check"
+ALTER TABLE t3 ADD COLUMN b domain6; --ok. because t3 is empty
+NOTICE:  rewriting table t3 for reason 2
+ALTER TABLE t3 ADD COLUMN c domain7; --ok. because t3 is empty
+NOTICE:  rewriting table t3 for reason 2
+ALTER TABLE t3 ADD COLUMN d domain5; --ok. because t3 is empty
+NOTICE:  rewriting table t3 for reason 2
+INSERT INTO t3(a,b,c,d) values(1,2, 3, default); --should fail
+ERROR:  value for domain domain5 violates check constraint "domain5_check"
+INSERT INTO t3(a,b,c,d) values(1,default, 3, 12); --should fail
+ERROR:  domain domain6 does not allow null values
+INSERT INTO t3(a,b,c,d) values(1,2, default, 12); --should fail
+ERROR:  value for domain domain7 violates check constraint "domain7_check"
+INSERT INTO t3(a,b,c,d) values(1,2, 3, 12); --ok
+DROP TABLE t3;
+CREATE TABLE t3(a int);
+INSERT INTO t3 VALUES(1),(2);
+ALTER TABLE t3 ADD COLUMN b domain5; --table rewrite, then fail
+NOTICE:  rewriting table t3 for reason 2
+ERROR:  value for domain domain5 violates check constraint "domain5_check"
+ALTER TABLE t3 ADD COLUMN b domain6; --table rewrite, then fail
+NOTICE:  rewriting table t3 for reason 2
+ERROR:  domain domain6 does not allow null values
+ALTER TABLE t3 ADD COLUMN b domain7; --table rewrite, then fail
+NOTICE:  rewriting table t3 for reason 2
+ERROR:  value for domain domain7 violates check constraint "domain7_check"
+ALTER TABLE t3 ADD COLUMN b domain5 default 12; --no table rewrite
+ALTER TABLE t3 ADD COLUMN c domain6 default 13; --no table rewrite
+ALTER TABLE t3 ADD COLUMN d domain7 default 14; --no table rewrite
+--no table rewrite we consider explicit column default expression, not domain default
+ALTER TABLE t3 ADD COLUMN e domain8 default 15;
+ALTER TABLE t3 ADD COLUMN f domain9; --no table rewrite
+SELECT attnum, attname, atthasmissing, atthasdef, attmissingval
+FROM pg_attribute
+WHERE attnum > 0 AND attrelid = 't3'::regclass AND attisdropped is false
+ORDER BY attnum;
+ attnum | attname | atthasmissing | atthasdef | attmissingval 
+--------+---------+---------------+-----------+---------------
+      1 | a       | f             | f         | 
+      2 | b       | t             | t         | {12}
+      3 | c       | t             | t         | {13}
+      4 | d       | t             | t         | {14}
+      5 | e       | t             | t         | {15}
+      6 | f       | t             | f         | {9}
+(6 rows)
+
+--table rewrite. we are applying domain volatile default expresion
+ALTER TABLE t3 ADD COLUMN g domain8;
+NOTICE:  rewriting table t3 for reason 2
+SELECT attnum, attname, atthasmissing, atthasdef, attmissingval
+FROM pg_attribute
+WHERE attnum > 0 AND attrelid = 't3'::regclass AND attisdropped is false
+ORDER BY attnum;
+ attnum | attname | atthasmissing | atthasdef | attmissingval 
+--------+---------+---------------+-----------+---------------
+      1 | a       | f             | f         | 
+      2 | b       | f             | t         | 
+      3 | c       | f             | t         | 
+      4 | d       | f             | t         | 
+      5 | e       | f             | t         | 
+      6 | f       | f             | f         | 
+      7 | g       | f             | f         | 
+(7 rows)
+
+SELECT a,b,c,d,e,f, g > 10 as f_ok FROM t3 ORDER BY a;
+ a | b  | c  | d  | e  | f | f_ok 
+---+----+----+----+----+---+------
+ 1 | 12 | 13 | 14 | 15 | 9 | t
+ 2 | 12 | 13 | 14 | 15 | 9 | t
+(2 rows)
+
 DROP TABLE t2;
+DROP TABLE t3;
 DROP DOMAIN domain1;
 DROP DOMAIN domain2;
 DROP DOMAIN domain3;
 DROP DOMAIN domain4;
+DROP DOMAIN domain5;
+DROP DOMAIN domain6;
+DROP DOMAIN domain7;
+DROP DOMAIN domain8;
+DROP DOMAIN domain9;
 DROP FUNCTION foo(INT);
 -- Fall back to full rewrite for volatile expressions
 CREATE TABLE T(pk INT NOT NULL PRIMARY KEY);
diff --git a/src/test/regress/sql/fast_default.sql b/src/test/regress/sql/fast_default.sql
index 068dd0bc8aa..e3139ce8b15 100644
--- a/src/test/regress/sql/fast_default.sql
+++ b/src/test/regress/sql/fast_default.sql
@@ -287,11 +287,68 @@ ORDER BY attnum;
 
 SELECT a, b, length(c) = 3 as c_ok, d, e >= 10 as e_ok FROM t2;
 
+---test fast default over domains with check constraint
+CREATE DOMAIN domain5 AS int check(value > 10) default 8;
+create domain domain6 as int not null;
+create domain domain7 as int check(value is not null);
+create domain domain8 as int check(value > 10) DEFAULT random(min=>10, max=>100);
+create domain domain9 as int check(value is not null) default 9;
+
+CREATE TABLE t3(a int);
+ALTER TABLE t3 ADD COLUMN b domain5 default 1; --error
+ALTER TABLE t3 ADD COLUMN b domain6 default NULL; --error
+ALTER TABLE t3 ADD COLUMN b domain6 default 11 + NULL; --error
+ALTER TABLE t3 ADD COLUMN b domain7 default 11 + NULL; --error
+
+ALTER TABLE t3 ADD COLUMN b domain6; --ok. because t3 is empty
+ALTER TABLE t3 ADD COLUMN c domain7; --ok. because t3 is empty
+ALTER TABLE t3 ADD COLUMN d domain5; --ok. because t3 is empty
+INSERT INTO t3(a,b,c,d) values(1,2, 3, default); --should fail
+INSERT INTO t3(a,b,c,d) values(1,default, 3, 12); --should fail
+INSERT INTO t3(a,b,c,d) values(1,2, default, 12); --should fail
+INSERT INTO t3(a,b,c,d) values(1,2, 3, 12); --ok
+
+DROP TABLE t3;
+CREATE TABLE t3(a int);
+INSERT INTO t3 VALUES(1),(2);
+
+ALTER TABLE t3 ADD COLUMN b domain5; --table rewrite, then fail
+ALTER TABLE t3 ADD COLUMN b domain6; --table rewrite, then fail
+ALTER TABLE t3 ADD COLUMN b domain7; --table rewrite, then fail
+
+ALTER TABLE t3 ADD COLUMN b domain5 default 12; --no table rewrite
+ALTER TABLE t3 ADD COLUMN c domain6 default 13; --no table rewrite
+ALTER TABLE t3 ADD COLUMN d domain7 default 14; --no table rewrite
+--no table rewrite we consider explicit column default expression, not domain default
+ALTER TABLE t3 ADD COLUMN e domain8 default 15;
+ALTER TABLE t3 ADD COLUMN f domain9; --no table rewrite
+
+SELECT attnum, attname, atthasmissing, atthasdef, attmissingval
+FROM pg_attribute
+WHERE attnum > 0 AND attrelid = 't3'::regclass AND attisdropped is false
+ORDER BY attnum;
+
+--table rewrite. we are applying domain volatile default expresion
+ALTER TABLE t3 ADD COLUMN g domain8;
+
+SELECT attnum, attname, atthasmissing, atthasdef, attmissingval
+FROM pg_attribute
+WHERE attnum > 0 AND attrelid = 't3'::regclass AND attisdropped is false
+ORDER BY attnum;
+
+SELECT a,b,c,d,e,f, g > 10 as f_ok FROM t3 ORDER BY a;
+
 DROP TABLE t2;
+DROP TABLE t3;
 DROP DOMAIN domain1;
 DROP DOMAIN domain2;
 DROP DOMAIN domain3;
 DROP DOMAIN domain4;
+DROP DOMAIN domain5;
+DROP DOMAIN domain6;
+DROP DOMAIN domain7;
+DROP DOMAIN domain8;
+DROP DOMAIN domain9;
 DROP FUNCTION foo(INT);
 
 -- Fall back to full rewrite for volatile expressions
-- 
2.34.1

